/*
 * File         : Fox.java
 * Author       : Guansong Zhang
 * Created      : Wed Nov  5 17:10:16 1997
 * Revision     : $Revision$
 * Updated      : $Date$
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1997
 */

import adlib.*;

class Fox {

  final static int P=2;
  final static int B=2;

  static Group root;

  public static void input1(Group p, Array4Float a) {
    // should be remaped from root
    for (int m=0; m<P; m++) {
      Location lm=a.range[0].location(m);     
      if (lm.at())
	for (int n=0; n<P; n++) {
	  Location ln=a.range[1].location(n);
	  if(ln.at())
	    for (int i=0; i<B; i++) {
	      Location li=a.range[2].location(i);
	      if(li.at())
		for (int j=0; j<B; j++) {
		  Location lj=a.range[3].location(j);
		  if(lj.at()) {
		    a.data[a.element(lm, ln, li, lj)]=6-(m+m+n+n+i+j);
		  } lj.ta();
		} li.ta();	  
	    } ln.ta();
	} lm.ta();
    }
  };

  public static void input0(Group p, Array4Float a) {
    // should be remaped from root
    for (int m=0; m<P; m++) {
      Location lm=a.range[0].location(m);     
      if (lm.at())
	for (int n=0; n<P; n++) {
	  Location ln=a.range[1].location(n);
	  if(ln.at())
	    for (int i=0; i<B; i++) {
	      Location li=a.range[2].location(i);
	      if(li.at())
		for (int j=0; j<B; j++) {
		  Location lj=a.range[3].location(j);
		  if(lj.at()) {
		    a.data[a.element(lm, ln, li, lj)]=m+m+n+n+i+j;
		  } lj.ta();
		} li.ta();	  
	    } ln.ta();
	} lm.ta();
    }
  };

  public static void output(Group p, Array4Float a) {
    // p has one single process as root

    Range clp = new Range(P);
    Range z= new Range(B);
    Array4Float res = new Array4Float (clp, clp, z, z, p);

    Adlib.remap(res, a);

    if (p.on()) {
      for (int m=0; m<P; m++) {
	Location lm=res.range[0].location(m);     
	for (int i=0; i<B; i++) {
	  Location li=res.range[2].location(i);
	  for (int n=0; n<P; n++) {
	    Location ln=res.range[1].location(n);
	    for (int j=0; j<B; j++) {
	      Location lj=res.range[3].location(j);
	      System.out.print((int)res.data[res.element(lm, ln, li, lj)]+"  ");
	    }
	  }
	  System.out.println("|");
	}
      }
    } p.no();    
  };

  public static void matmul(Array2Float c, Array2Float a, Array2Float b) {
    for (int i=0; i<B; i++)
      for (int j=0; j<B; j++)
	for (int k=0; k<B; k++)
	  c.data[c.element(c.range[0].location(i), c.range[1].location(j))]+=
	    a.data[a.element(a.range[0].location(i), a.range[1].location(k))]*
	    b.data[b.element(b.range[0].location(k), b.range[1].location(j))];
  }

  public static void main(String argv[]) {
    Adlib.Initialize(argv);

    int[] grid = {P, P};
    Procs p=new Procs(grid);

    Range x=p.dim(0);
    Range y=p.dim(1);

    root=p.restrict(x.location(0)).restrict(y.location(0));
      
    Range z=new Range(B);

    if(p.on()) {
      Array4Float a = new Array4Float(x,y,z,z,p);
      input0(root, a);
      output(root, a);
                  
      Array4Float b = new Array4Float(x,y,z,z,p);
      input1(root, b);
      output(root, b);

      Array4Float tmp = new Array4Float(x,y,z,z,p);

      Array4Float c = new Array4Float(x,y,z,z,p);

      for (int k=0; k<P; k++) {
	x.over(); {
	  int m=x.crd(); // coordinate in the range;
	  Location i=x.location(m);
	  Array2Float sub = new Array2Float(z,z,p.restrict(i));
	    
	  //Broadcast submatrix in 'a' ...
	  Adlib.remap(sub, a.section(i, y.location((m+k)%P), z, z));

	  y.over(); {
	    int n=y.crd();
	    Location j=y.location(n);

	    //Local matrix multiprlication ...
	    matmul(c.section(i, j, z, z), sub, b.section(i, j, z, z));
	  } y.revo();
	} x.revo();
	  
	//Cyclic shift 'b' in 'y' dimension ...
	Adlib.shift(tmp, b, 1, 0, 0); // dst, src, shift, dim, mode;
	Adlib.copy(b, tmp);
      }

      output(root, c);
    } p.no();

    Adlib.Finalize();
    System.exit(0);
  }
}
